// noinspection JSUnusedGlobalSymbols

/* Overwritten protection. */

let { util } = global;

/**
 * @param {ScriptRuntime} scriptRuntime
 * @param {org.mozilla.javascript.Scriptable | global} scope
 * @return {Internal.Automator}
 */
module.exports = function (scriptRuntime, scope) {

    const Path = android.graphics.Path;
    const Rect = android.graphics.Rect;
    const GestureDescription = android.accessibilityservice.GestureDescription;
    const AccessibilityTool = org.autojs.autojs.core.accessibility.AccessibilityTool;

    const AccessibilityBridge = org.autojs.autojs.core.accessibility.AccessibilityBridge;

    /** @type {org.autojs.autojs.core.accessibility.SimpleActionAutomator} */
    const rtAutomator = scriptRuntime.automator;

    /** @type {org.autojs.autojs.core.accessibility.AccessibilityBridge} */
    const a11yBridge = scriptRuntime.accessibilityBridge;

    const a11yTool = new AccessibilityTool(context);

    let _ = {
        Auto: (/* @IIFE */ () => {
            /**
             * @implements Internal.Auto
             */
            const Auto = function () {
                return Object.defineProperties(function __(mode, isForcibleRestart) {
                    if (isNullish(arguments[0])) {
                        return __(false);
                    }
                    if (typeof arguments[0] === 'boolean') {
                        a11yBridge.ensureServiceStarted(arguments[0]);
                        return;
                    }
                    if (typeof arguments[0] === 'string') {
                        auto.setMode(arguments[0]);
                    }
                    if (typeof arguments[1] === 'boolean') {
                        a11yBridge.ensureServiceStarted(arguments[1]);
                    }
                }, Object.getOwnPropertyDescriptors(Auto.prototype));
            };

            Auto.prototype = {
                constructor: Auto,
                toString() {
                    return 'auto';
                },
                get service() {
                    return a11yBridge.getService();
                },
                get services() {
                    return a11yTool.getServices();
                },
                get windows() {
                    return this.service === null ? [] : util.java.toJsArray(this.service.getWindows(), true);
                },
                get root() {
                    let root = a11yBridge.getRootInCurrentWindow();
                    return root ? UiObject.createRoot(root) : null;
                },
                get rootInActiveWindow() {
                    let root = a11yBridge.getRootInActiveWindow();
                    return root ? UiObject.createRoot(root) : null;
                },
                get windowRoots() {
                    return util.java.toJsArray(a11yBridge.windowRoots(), false)
                        .map(root => UiObject.createRoot(root));
                },
                start() {
                    return a11yTool.startService();
                },
                stop() {
                    return a11yTool.stopService();
                },
                isRunning() {
                    return a11yTool.isServiceRunning();
                },
                exists() {
                    return a11yTool.serviceExists();
                },
                stateListener(listener) {
                    return a11yBridge.setAccessibilityListener(listener);
                },
                registerEvent(name, listener) {
                    return rtAutomator.registerEvent(name, listener);
                },
                /** @deprecated */
                registerEvents(name, listener) {
                    return rtAutomator.registerEvent(name, listener);
                },
                removeEvent(name) {
                    return rtAutomator.removeEvent(name);
                },
                /** @deprecated */
                removeEvents(name) {
                    return rtAutomator.removeEvent(name);
                },
                waitFor(timeout) {
                    automator.waitForService(timeout);
                },
                setMode(modeStr) {
                    if (typeof modeStr !== 'string') {
                        throw TypeError('Mode should be a string for auto.setMode()');
                    }
                    let mode = _.modes[modeStr];
                    if (mode === undefined) {
                        throw Error(`Unknown mode for auto.setMode(): ${modeStr}`);
                    }
                    a11yBridge.setMode(mode);
                },
                setFlags(flags) {
                    let flagStrings;
                    if (Array.isArray(flags)) {
                        flagStrings = flags;
                    } else if (typeof flags === 'string') {
                        flagStrings = [ flags ];
                    } else {
                        throw TypeError(`Unknown flags: ${flags}`);
                    }
                    let flagsInt = 0;
                    flagStrings.forEach((s) => {
                        let flag = _.flagsMap[s];
                        if (flag === undefined) {
                            throw Error(`Unknown flag for auto.setFlags(): ${flag}`);
                        }
                        flagsInt |= flag;
                    });
                    a11yBridge.setFlags(flagsInt);
                },
                /**
                 * @param filter {(info: android.view.accessibility.AccessibilityWindowInfo) => boolean}
                 */
                setWindowFilter(filter) {
                    a11yBridge.setWindowFilter(new AccessibilityBridge.WindowFilter({ filter }));
                },
                launchSettings() {
                    a11yTool.launchSettings();
                },
                clearCache() {
                    return a11yTool.clearCache();
                },
            };

            return Auto;
        })(),
        Automator: (/* @IIFE */ () => {
            /**
             * @implements Internal.Automator
             */
            const Automator = function () {
                /* Empty body. */
            };

            Automator.prototype = {
                constructor: Automator,
                toString() {
                    return 'automator';
                },
                click() {
                    if (arguments.length === 2) {
                        let [ x, y ] = arguments;
                        if (typeof x === 'number' && typeof y === 'number') {
                            return rtAutomator.click(x, y);
                        }
                    }
                    let target = arguments[0];
                    if (target instanceof Rect) {
                        return this.click(target.centerX(), target.centerY());
                    }
                    if (target instanceof UiObject) {
                        return target.clickable() ? target.click() : this.click(target.bounds());
                    }
                    if (Array.isArray(target)) {
                        let [ x, y ] = target;
                        if (typeof x === 'number') {
                            if (typeof y === 'number') {
                                return this.click(x, y);
                            }
                            if (typeof y === 'undefined') {
                                return this.click(x, x);
                            }
                        }
                    }
                    return _.performAction(function (target) {
                        return rtAutomator.click(target);
                    }, arguments);
                },
                longClick() {
                    if (arguments.length === 2) {
                        let [ x, y ] = arguments;
                        if (typeof x === 'number' && typeof y === 'number') {
                            return rtAutomator.longClick(x, y);
                        }
                    }
                    return _.performAction(function (target) {
                        return rtAutomator.longClick(target);
                    }, arguments);
                },
                press(x, y, duration) {
                    if (arguments.length === 3) {
                        return rtAutomator.press(x, y, duration);
                    }
                    if (arguments.length === 2) {
                        if (typeof arguments[0] === 'number' && Array.isArray(arguments[1])) {
                            let [ x, y ] = arguments[1];
                            return rtAutomator.press(x, y, arguments[0]);
                        }
                        if (Array.isArray(arguments[0]) && typeof arguments[1] === 'number') {
                            let [ x, y ] = arguments[0];
                            return rtAutomator.press(x, y, arguments[1]);
                        }
                    }
                    throw TypeError(`Invalid arguments for automator.press`);
                },
                swipe(x1, y1, x2, y2, duration) {
                    if (arguments.length === 5) {
                        return rtAutomator.swipe(x1, y1, x2, y2, duration);
                    }
                    if (arguments.length === 2) {
                        if (typeof arguments[0] === 'number' && Array.isArray(arguments[1])) {
                            let [ [ xA, yA ], [ xB, yB ] ] = _.toPointsGroup(arguments[1]);
                            return rtAutomator.swipe(xA, yA, xB, yB, arguments[0]);
                        }
                        if (Array.isArray(arguments[0]) && typeof arguments[1] === 'number') {
                            let [ [ xA, yA ], [ xB, yB ] ] = _.toPointsGroup(arguments[0]);
                            return rtAutomator.swipe(xA, yA, xB, yB, arguments[1]);
                        }
                    }
                    if (arguments.length === 3) {
                        if (typeof arguments[0] === 'number' && Array.isArray(arguments[1]) && Array.isArray(arguments[2])) {
                            let [ [ xA, yA ], [ xB, yB ] ] = _.toPointsGroup(Array.from(arguments).slice(1));
                            return rtAutomator.swipe(xA, yA, xB, yB, arguments[0]);
                        }
                        if (Array.isArray(arguments[0]) && Array.isArray(arguments[1]) && typeof arguments[2] === 'number') {
                            let [ [ xA, yA ], [ xB, yB ] ] = _.toPointsGroup(Array.from(arguments).slice(0, arguments.length - 1));
                            return rtAutomator.swipe(xA, yA, xB, yB, arguments[2]);
                        }
                    }
                    throw TypeError(`Invalid arguments for automator.swipe`);
                },
                gesture(duration, vararg$point) {
                    return rtAutomator.gesture(0, duration, _.toPointsGroup(Array.from(arguments).slice(1)));
                },
                gestureAsync(duration, vararg$point) {
                    if (arguments.length < 1) {
                        return;
                    }
                    let start = 0;
                    let last = arguments[arguments.length - 1];
                    if (_.isGestureResultCallbackLike(last)) {
                        rtAutomator.gestureAsync(start, duration, _.toPointsGroup(Array.from(arguments).slice(1, -1)), _.toGestureResultCallback(last));
                    } else {
                        rtAutomator.gestureAsync(start, duration, _.toPointsGroup(Array.from(arguments).slice(1)));
                    }
                },
                gestures(vararg$stoke) {
                    return rtAutomator.gestures(_.toStrokes(arguments));
                },
                gesturesAsync(vararg$stroke) {
                    if (arguments.length < 1) {
                        return;
                    }
                    let last = arguments[arguments.length - 1];
                    if (_.isGestureResultCallbackLike(last)) {
                        rtAutomator.gesturesAsync(_.toStrokes(Array.from(arguments).slice(0, -1)), _.toGestureResultCallback(last));
                    } else {
                        rtAutomator.gesturesAsync(_.toStrokes(arguments));
                    }
                },
                isServiceRunning() {
                    return rtAutomator.isServiceRunning();
                },
                ensureService() {
                    rtAutomator.ensureService();
                },
                waitForService(timeout) {
                    a11yTool.startServiceAndWaitFor(_.parseNumber(timeout, -1));
                },
                input() {
                    if (arguments.length === 2) {
                        let [ index, text ] = arguments;
                        return rtAutomator.appendText(rtAutomator.editable(index), text);
                    } else {
                        let [ text ] = arguments;
                        return rtAutomator.appendText(rtAutomator.editable(-1), text);
                    }
                },
                scrollDown(index) {
                    if (typeof index === 'number') {
                        return rtAutomator.scrollForward(index);
                    }
                    if (arguments.length === 0) {
                        return rtAutomator.scrollMaxForward();
                    }
                    return _.performAction(function (target) {
                        return rtAutomator.scrollDown(target);
                    }, arguments);
                },
                scrollUp(index) {
                    if (typeof index === 'number') {
                        return rtAutomator.scrollBackward(index);
                    }
                    if (arguments.length === 0) {
                        return rtAutomator.scrollMaxBackward();
                    }
                    return _.performAction(function (target) {
                        return rtAutomator.scrollUp(target);
                    }, arguments);
                },
                setText() {
                    if (arguments.length === 2) {
                        let [ index, text ] = arguments;
                        return rtAutomator.setText(rtAutomator.editable(index), text);
                    } else {
                        let [ text ] = arguments;
                        return rtAutomator.setText(rtAutomator.editable(-1), text);
                    }
                },
                captureScreen() {
                    return ResultAdapter.wait(rtAutomator.captureScreen());
                },
                lockScreen() {
                    return rtAutomator.lockScreen();
                },
                takeScreenshot() {
                    return rtAutomator.takeScreenshot();
                },
                headsethook() {
                    return rtAutomator.headsethook();
                },
                accessibilityButton() {
                    return rtAutomator.accessibilityButton();
                },
                accessibilityButtonChooser() {
                    return rtAutomator.accessibilityButtonChooser();
                },
                accessibilityShortcut() {
                    return rtAutomator.accessibilityShortcut();
                },
                accessibilityAllApps() {
                    return rtAutomator.accessibilityAllApps();
                },
                dismissNotificationShade() {
                    return rtAutomator.dismissNotificationShade();
                },
            };

            return Automator;
        })(),
        modes: {
            normal: AccessibilityBridge.MODE_NORMAL,
            fast: AccessibilityBridge.MODE_FAST,
        },
        flagsMap: {
            findOnUiThread: AccessibilityBridge.FLAG_FIND_ON_UI_THREAD,
            useUsageStats: AccessibilityBridge.FLAG_USE_USAGE_STATS,
            useShell: AccessibilityBridge.FLAG_USE_SHELL,
        },
        /**
         * @template {boolean} T
         * @param {(target: org.autojs.autojs.core.automator.action.ActionTarget) => T} action
         * @param {IArguments} args
         * @return {T}
         */
        performAction(action, args) {
            if (args.length === 4) {
                let [ left, top, right, bottom ] = args;
                return action(rtAutomator.bounds(left, top, right, bottom));
            }
            if (args.length === 2) {
                let [ text, index ] = args;
                return action(rtAutomator.text(text, index));
            }
            let [ text ] = args;
            return action(rtAutomator.text(text, -1));
        },
        toStrokes(argsList) {
            let screenMetrics = scriptRuntime.getScreenMetrics();
            let strokes = java.lang.reflect.Array.newInstance(GestureDescription.StrokeDescription, argsList.length);

            for (let i = 0; i < argsList.length; i += 1) {
                let args = argsList[i];
                let startTime, durationIndex, pointsIndex;
                if (typeof args[1] /* duration */ === 'number') {
                    /* arguments: [startTime, duration, points[]] */
                    startTime = args[0];
                    durationIndex = 1;
                    pointsIndex = 2;
                } else {
                    /* arguments: [duration, points[]] */
                    startTime = 0; // default value
                    durationIndex = 0;
                    pointsIndex = 1;
                }
                let path = new Path();
                let [ x, y ] = args[pointsIndex];
                path.moveTo(screenMetrics.scaleX(x), screenMetrics.scaleY(y));
                for (let j = pointsIndex + 1; j < args.length; j += 1) {
                    let [ x, y ] = args[j];
                    path.lineTo(screenMetrics.scaleX(x), screenMetrics.scaleY(y));
                }
                strokes[i] = new GestureDescription.StrokeDescription(path, startTime, args[durationIndex]);
            }

            return strokes;
        },
        /**
         * @param {((isComplete: boolean) => void) | {
         *     onCompleted?(gestureDescription: android.accessibilityservice.GestureDescription): void;
         *     onCancelled?(gestureDescription: android.accessibilityservice.GestureDescription): void;
         * }} callback
         * @returns {GestureResultCallback|null}
         */
        toGestureResultCallback(callback) {
            return !this.isGestureResultCallbackLike(callback) ? null : new JavaAdapter(GestureResultCallback, ({
                onCompleted(gestureDescription) {
                    if (typeof callback === 'function') {
                        callback(true);
                    } else if (typeof callback === 'object' && typeof callback.onCompleted === 'function') {
                        callback.onCompleted(gestureDescription);
                    }
                },
                onCancelled(gestureDescription) {
                    if (typeof callback === 'function') {
                        callback(false);
                    } else if (typeof callback === 'object' && typeof callback.onCancelled === 'function') {
                        callback.onCancelled(gestureDescription);
                    }
                },
            }));
        },
        /**
         * @param {((isComplete: boolean) => void) | {
         *     onCompleted?(gestureDescription: android.accessibilityservice.GestureDescription): void;
         *     onCancelled?(gestureDescription: android.accessibilityservice.GestureDescription): void;
         * }} o
         * @returns {boolean}
         */
        isGestureResultCallbackLike(o) {
            return typeof o === 'function'
                || typeof o === 'object' && (typeof o.onCompleted === 'function' || typeof o.onCancelled === 'function');
        },
        /**
         * @param {*[]} args
         * @returns {[number, number][]}
         */
        toPointsGroup(args) {
            /** @type {[number, number][]} */
            let group = [];
            let argsArray = arguments.length === 1 && Array.isArray(args) ? args : Array.from(arguments);
            if (argsArray.length > 0 && argsArray.every(x => typeof x === 'number')) {
                if (argsArray.length % 2 !== 0) {
                    throw TypeError(`Arguments cannot be converted as points as the number of them must be even`);
                }
                for (let g = 0; g < argsArray.length / 2; g += 1) {
                    let tmp = [];
                    for (let i = 0; i < 2; i += 1) {
                        tmp[i] = argsArray[g * 2 + i];
                    }
                    group.push(tmp);
                }
                return group;
            }
            for (let i = 0; i < argsArray.length; i += 1) {
                let o = argsArray[i];
                if (!Array.isArray(o)) {
                    throw TypeError(`Argument "${o}" cannot be converted to a point as it must be an Array`);
                }
                if (o.length < 1) {
                    throw TypeError(`Argument "[]" cannot be converted to a point as it has no elements`);
                }
                if (o.some(a => Array.isArray(a))) {
                    if (o.every(a => Array.isArray(a))) {
                        group = group.concat(_.toPointsGroup(o));
                        continue;
                    }
                    let filtered = o.filter(x => !Array.isArray(x));
                    if (filtered.length <= 1) {
                        throw TypeError(`Element "${filtered[0]}" must be an Array`);
                    }
                    throw TypeError(`Elements "${filtered.join('", "')}" must be all Arrays`);
                }
                o.some((x) => {
                    if (!isNaN(x) && !isNullish(x) && typeof x !== 'boolean') {
                        return true;
                    }
                    throw TypeError(`Element "${x}" must be number-like`);
                });
                if (o.length === 1) {
                    group.push(Array(2).fill(Number(o[0])));
                    continue;
                }
                if (o.length === 2) {
                    group.push(o.map(n => Number(n)));
                    continue;
                }
                throw TypeError(`Argument "[${o}]" cannot be converted to a point as it has too much elements`);
            }
            return group;
        },
        /**
         * @param {any} num
         * @param {number|function():number} [def=0]
         * @returns {number}
         */
        parseNumber(num, def) {
            return typeof num === 'number' ? num : typeof def === 'function' ? def() : def || 0;
        },
        scopeAugment() {
            /**
             * @type {(keyof Internal.Automator)[]}
             */
            let methods = [
                'click', 'longClick', 'press', 'swipe',
                'gesture', 'gestures', 'gestureAsync', 'gesturesAsync',
                'scrollDown', 'scrollUp', 'input', 'setText',
            ];
            __asGlobal__(automator, methods, scope);

            /**
             * @type {(keyof org.autojs.autojs.core.accessibility.SimpleActionAutomator)[]}
             */
            let methodsRt = [
                'back', 'home', 'powerDialog', 'notifications',
                'quickSettings', 'recents', 'splitScreen',
            ];
            __asGlobal__(rtAutomator, methodsRt, scope);

            /**
             * @Caution by SuperMonster003 on Apr 23, 2022.
             * Use 'bind' or 'assign' will lose appended properties.
             *
             * @example
             * let f = function () {}; f.code = 1;
             * let g = f; console.log(g.code); // 1
             * let h = f.bind({}); console.log(h.code); // undefined
             * let o = {}; Object.assign(o, {f}); console.log(o.code); // undefined
             */
            scope.auto = auto;
        },
    };

    /**
     * @type {Internal.Auto}
     */
    const auto = new _.Auto();

    /**
     * @type {Internal.Automator}
     */
    const automator = new _.Automator();

    _.scopeAugment();

    return automator;
};