// noinspection JSUnusedLocalSymbols,UnnecessaryLocalVariableJS

/**
 * @param {ScriptRuntime} scriptRuntime
 * @param {org.mozilla.javascript.Scriptable | global} scope
 * @return {Internal.S13n}
 */
module.exports = function (scriptRuntime, scope) {

    const TimeUtils = org.autojs.autojs.util.TimeUtils;

    let _ = {
        rex: {
            nanoSeconds: /^NANO(SECONDS?|S)?$/i,
            microSeconds: /^MICRO(SECONDS?|S)?$/i,
            milliSeconds: /^(MILLI(SECONDS?|S)?|MS)$/i,
            seconds: /^(S(ECONDS?)?|SECS?)$/i,
            minutes: /^(M(INUTES?)?|MINS?)$/i,
            hours: /^(H(OURS?)?|HRS?)$/i,
            days: /^D(AYS?)?$/i,
        },
        S13n: (/* @IIFE */ () => {
            /**
             * @extends Internal.S13n
             */
            const S13n = function () {
                /* Empty body. */
            };

            S13n.prototype = {
                constructor: S13n,
                color(o) {
                    if (typeof o === 'string') {
                        // ColorHex | ColorName
                        return Color(o).toInt();
                    }
                    if (typeof o === 'number') {
                        // ColorInt
                        return Color(o).toInt(); /* For numbers like 0xRRGGBB. */
                    }
                    if (o instanceof Color) {
                        return o.toInt();
                    }
                    if (o instanceof org.autojs.autojs.theme.ThemeColor) {
                        return o.getColorPrimary();
                    }
                    throw TypeError(`Failed to make "{value: ${o}, species: ${species(o)}})" a color being`);
                },
                throwable(o) {
                    if (o instanceof java.lang.Throwable) {
                        return o;
                    }
                    if (typeof o === 'string') {
                        return new java.lang.Exception(o);
                    }
                    if (util.getClassName(o) === 'org.mozilla.javascript.NativeError') {
                        return o.rhinoException || o.javaException || this.throwable(o.message);
                    }
                    throw TypeError(`Failed to make "{value: ${o}, species: ${species(o)}})" a throwable being`);
                },
                point(o) {

                    // TODO by SuperMonster003 on Jan 14, 2024.
                    //  ! Support s13n.point(0.2, 0.8).

                    if (o instanceof android.graphics.Point) {
                        return o;
                    }
                    if (o instanceof org.opencv.core.Point) {
                        return new android.graphics.Point(o.x, o.y);
                    }
                    if (Array.isArray(o)) {
                        let [ x, y ] = o;
                        return new android.graphics.Point(x, y);
                    }
                    if (typeof o === 'number') {
                        let x = o;
                        let y = Number(arguments[1]);
                        if (isNaN(y)) {
                            y = o;
                        }
                        return new android.graphics.Point(x, y);
                    }
                    if (isObject(o)) {
                        let x = Number(o.x);
                        let y = Number(o.y);
                        if (!isNaN(x) && !isNaN(y)) {
                            return new android.graphics.Point(x, y);
                        }
                    }
                    throw TypeError(`Failed to make "{value: ${o}, species: ${species(o)}})" a point being`);
                },
                time(source, fromUnit, toUnit) {
                    let niceSource = source;
                    if (Number.isNaN(source)) {
                        return NaN;
                    }
                    if (source instanceof Date) {
                        return this.time.apply(this, [ source.getTime() ].concat(Array.from(arguments).slice(1)));
                    }
                    if (typeof source === 'string') {
                        niceSource = source.replace(/\s+(?!\d)|(\D|^)\s+/g, '$1');
                        // TODO by SuperMonster003 on Jan 14, 2024.
                        //  ! Support s13n.time('00:02:10', unit).
                        let split = niceSource.split(/\D+/).filter(o => o !== '');
                        if (split.length > 6 || split.length === 0) {
                            return NaN;
                        }
                        if (split.length === 6) {
                            let [ y, m, d, hh, mm, ss ] = split;
                            let src = new Date(`${y}/${m}/${d} ${hh}:${mm}:${ss}`).getTime();
                            return this.time.apply(this, [ src ].concat(Array.from(arguments).slice(1)));
                        }
                    }
                    if (isObject(fromUnit)) {
                        let options = fromUnit;
                        return this.time(niceSource, options.fromUnit, options.toUnit);
                    }
                    let num = Number(niceSource);
                    if (isNaN(num)) {
                        throw TypeError(`Failed to make "{value: ${niceSource}, species: ${species(niceSource)}})" a time being`);
                    }
                    let niceFromUnit = isNullish(fromUnit) ? TimeUnit.MILLISECONDS : fromUnit;
                    let niceToUnit = isNullish(toUnit) ? TimeUnit.MILLISECONDS : toUnit;
                    if (isNullish(niceFromUnit)) {
                        return num;
                    }
                    return TimeUtils[_.getTimeUnitConversionFunctionName(niceToUnit)](_.getTimeUnitSourceObject(niceFromUnit), num);
                },
            };

            return S13n;
        })(),
        /**
         * @param {string | java.util.concurrent.TimeUnit} unit
         * @returns {java.util.concurrent.TimeUnit}
         */
        getTimeUnitSourceObject(unit) {
            if (unit instanceof TimeUnit) {
                return unit;
            }
            if (typeof unit === 'string') {
                if (_.rex.nanoSeconds.test(unit)) {
                    return TimeUnit.NANOSECONDS;
                }
                if (_.rex.microSeconds.test(unit)) {
                    return TimeUnit.MICROSECONDS;
                }
                if (_.rex.milliSeconds.test(unit)) {
                    return TimeUnit.MILLISECONDS;
                }
                if (_.rex.seconds.test(unit)) {
                    return TimeUnit.SECONDS;
                }
                if (_.rex.minutes.test(unit)) {
                    return TimeUnit.MINUTES;
                }
                if (_.rex.hours.test(unit)) {
                    return TimeUnit.HOURS;
                }
                if (_.rex.days.test(unit)) {
                    return TimeUnit.DAYS;
                }
            }
            throw TypeError(`Failed to make "{value: ${unit}, species: ${species(unit)}})" a time unit being`);
        },
        /**
         * @param {string | java.util.concurrent.TimeUnit} unit
         * @returns {string}
         */
        getTimeUnitConversionFunctionName(unit) {
            if (unit instanceof TimeUnit) {
                switch (unit) {
                    case TimeUnit.NANOSECONDS:
                        return 'toNanos';
                    case TimeUnit.MICROSECONDS:
                        return 'toMicros';
                    case TimeUnit.MILLISECONDS:
                        return 'toMillis';
                    case TimeUnit.SECONDS:
                        return 'toSeconds';
                    case TimeUnit.MINUTES:
                        return 'toMinutes';
                    case TimeUnit.HOURS:
                        return 'toHours';
                    case TimeUnit.DAYS:
                        return 'toDays';
                }
            }
            if (typeof unit === 'string') {
                if (_.rex.nanoSeconds.test(unit)) {
                    return 'toNanos';
                }
                if (_.rex.microSeconds.test(unit)) {
                    return 'toMicros';
                }
                if (_.rex.milliSeconds.test(unit)) {
                    return 'toMillis';
                }
                if (_.rex.seconds.test(unit)) {
                    return 'toSeconds';
                }
                if (_.rex.minutes.test(unit)) {
                    return 'toMinutes';
                }
                if (_.rex.hours.test(unit)) {
                    return 'toHours';
                }
                if (_.rex.days.test(unit)) {
                    return 'toDays';
                }
            }
            throw TypeError(`Failed to make "{value: ${unit}, species: ${species(unit)}})" a time unit conversion function name`);
        },
    };

    /**
     * @type {Internal.S13n}
     */
    const s13n = new _.S13n();

    return s13n;
};